/** 
 * DevExtreme (animation/frame.js)
 * Version: 16.1.5
 * Build date: Mon Jul 25 2016
 *
 * Copyright (c) 2012 - 2016 Developer Express Inc. ALL RIGHTS RESERVED
 * EULA: https://www.devexpress.com/Support/EULAs/DevExtreme.xml
 * MAY BE USED WITH: DevExtreme Complete, DevExtreme Web, DevExtreme Mobile
 */
"use strict";
define(function(require, exports, module) {
    var $ = require("jquery"),
        FRAME_ANIMATION_STEP_TIME = 1e3 / 60,
        requestAnimationFrame = function(callback) {
            return this.setTimeout(callback, FRAME_ANIMATION_STEP_TIME)
        },
        cancelAnimationFrame = function(requestID) {
            this.clearTimeout(requestID)
        },
        nativeRequestAnimationFrame = window.requestAnimationFrame || window.webkitRequestAnimationFrame || window.mozRequestAnimationFrame || window.oRequestAnimationFrame || window.msRequestAnimationFrame,
        nativeCancelAnimationFrame = window.cancelAnimationFrame || window.webkitCancelAnimationFrame || window.mozCancelAnimationFrame || window.oCancelAnimationFrame || window.msCancelAnimationFrame;
    if (nativeRequestAnimationFrame && nativeCancelAnimationFrame) {
        requestAnimationFrame = nativeRequestAnimationFrame;
        cancelAnimationFrame = nativeCancelAnimationFrame
    }
    if (nativeRequestAnimationFrame && !nativeCancelAnimationFrame) {
        var canceledRequests = {};
        requestAnimationFrame = function(callback) {
            var requestId = nativeRequestAnimationFrame.call(window, function() {
                try {
                    if (requestId in canceledRequests) {
                        return
                    }
                    callback.apply(this, arguments)
                } finally {
                    delete canceledRequests[requestId]
                }
            });
            return requestId
        };
        cancelAnimationFrame = function(requestId) {
            canceledRequests[requestId] = true
        }
    }
    requestAnimationFrame = $.proxy(requestAnimationFrame, window);
    cancelAnimationFrame = $.proxy(cancelAnimationFrame, window);
    exports.requestAnimationFrame = requestAnimationFrame;
    exports.cancelAnimationFrame = cancelAnimationFrame
});
