/** 
 * DevExtreme (viz/axes/xy_axes.js)
 * Version: 16.1.5
 * Build date: Mon Jul 25 2016
 *
 * Copyright (c) 2012 - 2016 Developer Express Inc. ALL RIGHTS RESERVED
 * EULA: https://www.devexpress.com/Support/EULAs/DevExtreme.xml
 * MAY BE USED WITH: DevExtreme Complete, DevExtreme Web
 */
"use strict";
define(function(require, exports, module) {
    var $ = require("jquery"),
        formatHelper = require("../../format_helper"),
        dateUtils = require("../../core/utils/date"),
        commonUtils = require("../../core/utils/common"),
        vizUtils = require("../core/utils"),
        _isDefined = commonUtils.isDefined,
        constants = require("./axes_constants"),
        _extend = $.extend,
        CANVAS_POSITION_PREFIX = constants.canvasPositionPrefix,
        TOP = constants.top,
        BOTTOM = constants.bottom,
        LEFT = constants.left,
        RIGHT = constants.right,
        CENTER = constants.center;
    var dateSetters = {
        millisecond: function(date) {
            date.setMilliseconds(0)
        },
        second: function(date) {
            date.setSeconds(0, 0)
        },
        minute: function(date) {
            date.setMinutes(0, 0, 0)
        },
        hour: function(date) {
            date.setHours(0, 0, 0, 0)
        },
        month: function(date) {
            date.setMonth(0);
            dateSetters.day(date)
        },
        quarter: function(date) {
            date.setMonth(dateUtils.getFirstQuarterMonth(date.getMonth()));
            dateSetters.day(date)
        }
    };
    dateSetters.week = dateSetters.day = function(date) {
        date.setDate(1);
        dateSetters.hour(date)
    };

    function getMarkerDate(date, tickInterval) {
        var markerDate = new Date(date.getTime()),
            setter = dateSetters[tickInterval];
        setter && setter(markerDate);
        return markerDate
    }
    module.exports = {
        linear: {
            measureLabels: function() {
                return this._tickManager.getMaxLabelParams()
            },
            getMarkerTrackers: function() {
                return this._markerTrackers
            },
            _prepareDatesDifferences: function(datesDifferences, tickInterval) {
                var dateUnitInterval, i;
                if ("week" === tickInterval) {
                    tickInterval = "day"
                }
                if ("quarter" === tickInterval) {
                    tickInterval = "month"
                }
                if (datesDifferences[tickInterval]) {
                    for (i = 0; i < dateUtils.dateUnitIntervals.length; i++) {
                        dateUnitInterval = dateUtils.dateUnitIntervals[i];
                        if (datesDifferences[dateUnitInterval]) {
                            datesDifferences[dateUnitInterval] = false;
                            datesDifferences.count--
                        }
                        if (dateUnitInterval === tickInterval) {
                            break
                        }
                    }
                }
            },
            _getSharpParam: function(opposite) {
                return this._isHorizontal ^ opposite ? "h" : "v"
            },
            _createAxisElement: function() {
                var axisCoord = this._axisPosition,
                    canvas = this._getCanvasStartEnd(),
                    points = this._isHorizontal ? [canvas.start, axisCoord, canvas.end, axisCoord] : [axisCoord, canvas.start, axisCoord, canvas.end];
                return this._renderer.path(points, "line")
            },
            _getTranslatedCoord: function(value, offset) {
                return this._translator.translate(value, offset)
            },
            _getCanvasStartEnd: function() {
                return {
                    start: this._translator.translateSpecialCase(constants.canvasPositionStart),
                    end: this._translator.translateSpecialCase(constants.canvasPositionEnd)
                }
            },
            _getScreenDelta: function() {
                return Math.abs(this._translator.translateSpecialCase(constants.canvasPositionStart) - this._translator.translateSpecialCase(constants.canvasPositionEnd))
            },
            _initAxisPositions: function() {
                var that = this,
                    position = that._options.position,
                    delta = 0;
                if (that.delta) {
                    delta = that.delta[position] || 0
                }
                that._axisPosition = that._additionalTranslator.translateSpecialCase(CANVAS_POSITION_PREFIX + position) + delta
            },
            _getTickCoord: function(tick) {
                var coords, corrections = {
                        top: -1,
                        middle: -.5,
                        bottom: 0,
                        left: -1,
                        center: -.5,
                        right: 0
                    },
                    tickCorrection = corrections[this._options.tickOrientation || "center"];
                if (_isDefined(tick.posX) && _isDefined(tick.posY)) {
                    coords = {
                        x1: tick.posX,
                        y1: tick.posY + tickCorrection * tick.length,
                        x2: tick.posX,
                        y2: tick.posY + tickCorrection * tick.length + tick.length
                    }
                } else {
                    coords = null
                }
                return coords
            },
            _drawTitle: function() {
                var that = this,
                    options = that._options,
                    titleOptions = options.title,
                    attr = {
                        opacity: titleOptions.opacity,
                        align: CENTER
                    };
                if (!titleOptions.text || !that._axisTitleGroup) {
                    return
                }
                that._title = that._renderer.text(titleOptions.text, 0, 0).css(vizUtils.patchFontOptions(titleOptions.font)).attr(attr).append(that._axisTitleGroup)
            },
            _drawDateMarker: function(dateMarker, options) {
                var labelPosX, labelPosY, textElement, textSize, textIndent, pathElement, that = this,
                    markerOptions = that._options.marker;
                if (null === options.x) {
                    return
                }
                if (!options.withoutStick) {
                    pathElement = that._renderer.path([options.x, options.y, options.x, options.y + markerOptions.separatorHeight], "line").attr({
                        "stroke-width": markerOptions.width,
                        stroke: markerOptions.color,
                        "stroke-opacity": markerOptions.opacity,
                        sharp: "h"
                    }).append(that._axisElementsGroup)
                }
                textElement = that._renderer.text(String(constants.formatLabel(dateMarker, options.labelFormat)), 0, 0).attr({
                    align: "left"
                }).css(vizUtils.patchFontOptions(markerOptions.label.font)).append(that._axisElementsGroup);
                textSize = textElement.getBBox();
                textIndent = markerOptions.width + markerOptions.textLeftIndent;
                labelPosX = this._translator.getBusinessRange().invert ? options.x - textIndent - textSize.width : options.x + textIndent;
                labelPosY = options.y + markerOptions.textTopIndent + textSize.height / 2;
                textElement.move(labelPosX, labelPosY);
                return {
                    labelStartPosX: labelPosX,
                    labelEndPosX: labelPosX + textSize.width,
                    path: pathElement,
                    text: textElement,
                    date: dateMarker,
                    dateMarkerStartPosX: options.x
                }
            },
            _disposeDateMarker: function(marker) {
                marker.path && marker.path.dispose();
                marker.path = null;
                marker.text.dispose();
                marker.text = null
            },
            _getDiff: function(currentValue, previousValue) {
                var datesDifferences = dateUtils.getDatesDifferences(previousValue, currentValue);
                this._prepareDatesDifferences(datesDifferences, this._dateUnitInterval);
                return datesDifferences
            },
            _drawDateMarkers: function() {
                var prevDateMarker, markersAreaTop, dateMarker, markerDate, diff, that = this,
                    options = that._options,
                    ticks = that._majorTicks,
                    boundaryTicks = that._boundaryTicks,
                    lastIndexOfBoundaryTicks = boundaryTicks.length - 1,
                    length = ticks.length,
                    dateMarkers = [],
                    i = 1;
                boundaryTicks[0] && ticks[0].value > boundaryTicks[0].value && ticks.unshift(boundaryTicks[0]);
                boundaryTicks[lastIndexOfBoundaryTicks] && ticks[length - 1].value < boundaryTicks[lastIndexOfBoundaryTicks].value && ticks.push(boundaryTicks[lastIndexOfBoundaryTicks]);
                length = ticks.length;
                if ("datetime" !== options.argumentType || "discrete" === options.type || length <= 1) {
                    return
                }
                markersAreaTop = that._axisPosition + this._axisElementsGroup.getBBox().height + options.label.indentFromAxis + options.marker.topIndent;
                that._dateUnitInterval = dateUtils.getDateUnitInterval(this._tickManager.getTickInterval());
                for (i; i < length; i++) {
                    diff = that._getDiff(ticks[i].value, ticks[i - 1].value);
                    if (diff.count > 0) {
                        markerDate = getMarkerDate(ticks[i].value, that._dateUnitInterval);
                        dateMarker = that._drawDateMarker(markerDate, {
                            x: that._translator.translate(markerDate),
                            y: markersAreaTop,
                            labelFormat: that._getLabelFormatOptions(formatHelper.getDateFormatByDifferences(diff))
                        });
                        if (dateMarker) {
                            if (that._checkMarkersPosition(dateMarker, prevDateMarker)) {
                                dateMarkers.push(dateMarker);
                                prevDateMarker = dateMarker
                            } else {
                                that._disposeDateMarker(dateMarker)
                            }
                        }
                    }
                }
                if (dateMarkers.length) {
                    dateMarker = that._drawDateMarker(ticks[0].value, {
                        x: that._translator.translate(ticks[0].value),
                        y: markersAreaTop,
                        labelFormat: that._getLabelFormatOptions(formatHelper.getDateFormatByDifferences(that._getDiff(ticks[0].value, dateMarkers[0].date))),
                        withoutStick: true
                    });
                    if (dateMarker) {
                        !that._checkMarkersPosition(dateMarker, dateMarkers[0]) && that._disposeDateMarker(dateMarker);
                        dateMarkers.unshift(dateMarker)
                    }
                }
                that._initializeMarkersTrackers(dateMarkers, that._axisElementsGroup, that._axisGroup.getBBox().width, markersAreaTop)
            },
            _initializeMarkersTrackers: function(dateMarkers, group, axisWidth, markersAreaTop) {
                var markerTracker, nextMarker, i, x, currentMarker, that = this,
                    separatorHeight = that._options.marker.separatorHeight,
                    renderer = that._renderer,
                    length = dateMarkers.length,
                    businessRange = this._translator.getBusinessRange();
                that._markerTrackers = [];
                for (i = 0; i < length; i++) {
                    currentMarker = dateMarkers[i];
                    nextMarker = dateMarkers[i + 1] || {
                        dateMarkerStartPosX: businessRange.invert ? this._translator.translateSpecialCase("canvas_position_end") : axisWidth,
                        date: businessRange.max
                    };
                    x = currentMarker.dateMarkerStartPosX;
                    markerTracker = renderer.path([x, markersAreaTop, x, markersAreaTop + separatorHeight, nextMarker.dateMarkerStartPosX, markersAreaTop + separatorHeight, nextMarker.dateMarkerStartPosX, markersAreaTop, x, markersAreaTop]).attr({
                        "stroke-width": 1,
                        stroke: "grey",
                        fill: "grey",
                        "fill-opacity": 1e-4,
                        "stroke-opacity": 1e-4
                    }).append(group);
                    markerTracker.data("range", {
                        startValue: currentMarker.date,
                        endValue: nextMarker.date
                    });
                    that._markerTrackers.push(markerTracker)
                }
            },
            _checkMarkersPosition: function(dateMarker, prevDateMarker) {
                return void 0 === prevDateMarker || dateMarker.labelStartPosX > prevDateMarker.labelEndPosX || dateMarker.labelEndPosX < prevDateMarker.labelStartPosX
            },
            _getLabelFormatOptions: function(formatString) {
                var that = this,
                    markerLabelOptions = that._markerLabelOptions;
                if (!markerLabelOptions) {
                    that._markerLabelOptions = markerLabelOptions = _extend(true, {}, that._options.marker.label)
                }
                if (!_isDefined(that._options.marker.label.format)) {
                    markerLabelOptions.format = formatString
                }
                return markerLabelOptions
            },
            _adjustConstantLineLabels: function() {
                var label, line, lineBox, linesOptions, labelOptions, box, x, y, i, paddingTopBottom, paddingLeftRight, labelVerticalAlignment, labelHorizontalAlignment, labelIsInside, labelHeight, labelWidth, that = this,
                    options = that._options,
                    isHorizontal = that._isHorizontal,
                    lines = that._constantLines,
                    labels = that._constantLineLabels,
                    padding = isHorizontal ? {
                        top: 0,
                        bottom: 0
                    } : {
                        left: 0,
                        right: 0
                    },
                    delta = 0;
                if (void 0 === labels && void 0 === lines) {
                    return
                }
                for (i = 0; i < labels.length; i++) {
                    x = y = 0;
                    linesOptions = options.constantLines[i];
                    paddingTopBottom = linesOptions.paddingTopBottom;
                    paddingLeftRight = linesOptions.paddingLeftRight;
                    labelOptions = linesOptions.label;
                    labelVerticalAlignment = labelOptions.verticalAlignment;
                    labelHorizontalAlignment = labelOptions.horizontalAlignment;
                    labelIsInside = "inside" === labelOptions.position;
                    label = labels[i];
                    if (null !== label) {
                        line = lines[i];
                        box = label.getBBox();
                        lineBox = line.getBBox();
                        labelHeight = box.height;
                        labelWidth = box.width;
                        if (isHorizontal) {
                            if (labelIsInside) {
                                if (labelHorizontalAlignment === LEFT) {
                                    x -= paddingLeftRight
                                } else {
                                    x += paddingLeftRight
                                }
                                switch (labelVerticalAlignment) {
                                    case CENTER:
                                        y += lineBox.y + lineBox.height / 2 - box.y - labelHeight / 2;
                                        break;
                                    case BOTTOM:
                                        y += lineBox.y + lineBox.height - box.y - labelHeight - paddingTopBottom;
                                        break;
                                    default:
                                        y += lineBox.y - box.y + paddingTopBottom
                                }
                            } else {
                                if (labelVerticalAlignment === BOTTOM) {
                                    delta = that.delta && that.delta[BOTTOM] || 0;
                                    y += paddingTopBottom - box.y + that._additionalTranslator.translateSpecialCase(CANVAS_POSITION_PREFIX + BOTTOM) + delta;
                                    if (padding[BOTTOM] < labelHeight + paddingTopBottom) {
                                        padding[BOTTOM] = labelHeight + paddingTopBottom
                                    }
                                } else {
                                    delta = that.delta && that.delta[TOP] || 0;
                                    y -= paddingTopBottom + box.y + labelHeight - that._additionalTranslator.translateSpecialCase(CANVAS_POSITION_PREFIX + TOP) - delta;
                                    if (padding[TOP] < paddingTopBottom + labelHeight) {
                                        padding[TOP] = paddingTopBottom + labelHeight
                                    }
                                }
                            }
                        } else {
                            if (labelIsInside) {
                                switch (labelHorizontalAlignment) {
                                    case CENTER:
                                        x += lineBox.x + lineBox.width / 2 - box.x - labelWidth / 2;
                                        break;
                                    case RIGHT:
                                        x -= paddingLeftRight;
                                        break;
                                    default:
                                        x += paddingLeftRight
                                }
                                if (labelVerticalAlignment === BOTTOM) {
                                    y += lineBox.y - box.y + paddingTopBottom
                                } else {
                                    y += lineBox.y - box.y - labelHeight - paddingTopBottom
                                }
                            } else {
                                y += lineBox.y + lineBox.height / 2 - box.y - labelHeight / 2;
                                if (labelHorizontalAlignment === RIGHT) {
                                    x += paddingLeftRight;
                                    if (padding[RIGHT] < paddingLeftRight + labelWidth) {
                                        padding[RIGHT] = paddingLeftRight + labelWidth
                                    }
                                } else {
                                    x -= paddingLeftRight;
                                    if (padding[LEFT] < paddingLeftRight + labelWidth) {
                                        padding[LEFT] = paddingLeftRight + labelWidth
                                    }
                                }
                            }
                        }
                        label.move(x, y)
                    }
                }
                that.padding = padding
            },
            _checkAlignmentConstantLineLabels: function(labelOptions) {
                var position = labelOptions.position,
                    verticalAlignment = (labelOptions.verticalAlignment || "").toLowerCase(),
                    horizontalAlignment = (labelOptions.horizontalAlignment || "").toLowerCase();
                if (this._isHorizontal) {
                    if ("outside" === position) {
                        verticalAlignment = verticalAlignment === BOTTOM ? BOTTOM : TOP;
                        horizontalAlignment = CENTER
                    } else {
                        verticalAlignment = verticalAlignment === CENTER ? CENTER : verticalAlignment === BOTTOM ? BOTTOM : TOP;
                        horizontalAlignment = horizontalAlignment === LEFT ? LEFT : RIGHT
                    }
                } else {
                    if ("outside" === position) {
                        verticalAlignment = CENTER;
                        horizontalAlignment = horizontalAlignment === LEFT ? LEFT : RIGHT
                    } else {
                        verticalAlignment = verticalAlignment === BOTTOM ? BOTTOM : TOP;
                        horizontalAlignment = horizontalAlignment === RIGHT ? RIGHT : horizontalAlignment === CENTER ? CENTER : LEFT
                    }
                }
                labelOptions.verticalAlignment = verticalAlignment;
                labelOptions.horizontalAlignment = horizontalAlignment
            },
            _getConstantLineLabelsCoords: function(value, lineLabelOptions) {
                var that = this,
                    additionalTranslator = that._additionalTranslator,
                    align = CENTER,
                    x = value,
                    y = value;
                if (that._isHorizontal) {
                    y = additionalTranslator.translateSpecialCase(CANVAS_POSITION_PREFIX + lineLabelOptions.verticalAlignment)
                } else {
                    x = additionalTranslator.translateSpecialCase(CANVAS_POSITION_PREFIX + lineLabelOptions.horizontalAlignment)
                }
                switch (lineLabelOptions.horizontalAlignment) {
                    case LEFT:
                        align = !that._isHorizontal && "inside" === lineLabelOptions.position ? LEFT : RIGHT;
                        break;
                    case CENTER:
                        align = CENTER;
                        break;
                    case RIGHT:
                        align = !that._isHorizontal && "inside" === lineLabelOptions.position ? RIGHT : LEFT
                }
                return {
                    x: x,
                    y: y,
                    align: align
                }
            },
            _getAdjustedStripLabelCoords: function(stripOptions, label, rect) {
                var x = 0,
                    y = 0,
                    horizontalAlignment = stripOptions.label.horizontalAlignment,
                    verticalAlignment = stripOptions.label.verticalAlignment,
                    box = label.getBBox(),
                    rectBox = rect.getBBox();
                if (horizontalAlignment === LEFT) {
                    x += stripOptions.paddingLeftRight
                } else {
                    if (horizontalAlignment === RIGHT) {
                        x -= stripOptions.paddingLeftRight
                    }
                }
                if (verticalAlignment === TOP) {
                    y += rectBox.y - box.y + stripOptions.paddingTopBottom
                } else {
                    if (verticalAlignment === CENTER) {
                        y += rectBox.y + rectBox.height / 2 - box.y - box.height / 2
                    } else {
                        if (verticalAlignment === BOTTOM) {
                            y -= stripOptions.paddingTopBottom
                        }
                    }
                }
                return {
                    x: x,
                    y: y
                }
            },
            _adjustTitle: function() {
                var boxGroup, boxTitle, params, heightTitle, noLabels, that = this,
                    options = that._options,
                    position = options.position,
                    title = that._title,
                    margin = options.title.margin,
                    centerPosition = that._translator.translateSpecialCase(CANVAS_POSITION_PREFIX + CENTER),
                    axisElementsGroup = that._axisElementsGroup,
                    axisPosition = that._axisPosition;
                if (!title || !axisElementsGroup) {
                    return
                }
                boxTitle = title.getBBox();
                boxGroup = axisElementsGroup.getBBox();
                noLabels = boxGroup.isEmpty;
                heightTitle = boxTitle.height;
                if (that._isHorizontal) {
                    if (position === BOTTOM) {
                        params = {
                            y: (noLabels ? axisPosition : boxGroup.y + boxGroup.height) - boxTitle.y + margin,
                            x: centerPosition
                        }
                    } else {
                        params = {
                            y: (noLabels ? axisPosition : boxGroup.y) - heightTitle - boxTitle.y - margin,
                            x: centerPosition
                        }
                    }
                } else {
                    if (position === LEFT) {
                        params = {
                            x: (noLabels ? axisPosition : boxGroup.x) - heightTitle - boxTitle.y - margin,
                            y: centerPosition
                        }
                    } else {
                        params = {
                            x: (noLabels ? axisPosition : boxGroup.x + boxGroup.width) + heightTitle + boxTitle.y + margin,
                            y: centerPosition
                        }
                    }
                    params.rotate = options.position === LEFT ? 270 : 90
                }
                title.attr(params)
            },
            coordsIn: function(x, y) {
                var rect = this.getBoundingRect();
                return x >= rect.x && x <= rect.x + rect.width && y >= rect.y && y <= rect.y + rect.height
            },
            _boundaryTicksVisibility: {
                min: true,
                max: true
            },
            _getOverlappingBehaviorOptions: function() {
                var that = this,
                    options = that._options,
                    getText = function() {
                        return ""
                    },
                    overlappingBehavior = options.label.overlappingBehavior ? _extend({}, options.label.overlappingBehavior) : null;
                if (overlappingBehavior) {
                    if (!that._isHorizontal) {
                        overlappingBehavior.mode = constants.validateOverlappingMode(overlappingBehavior.mode)
                    }
                    if ("rotate" !== overlappingBehavior.mode) {
                        overlappingBehavior.rotationAngle = 0
                    }
                }
                if (!that._translator.getBusinessRange().stubData) {
                    getText = function(value, labelOptions) {
                        return constants.formatLabel(value, labelOptions, {
                            min: options.min,
                            max: options.max
                        })
                    }
                }
                return {
                    hasLabelFormat: that._hasLabelFormat,
                    labelOptions: options.label,
                    isMarkersVisible: "discrete" === options.type ? false : options.marker.visible,
                    overlappingBehavior: overlappingBehavior,
                    isHorizontal: that._isHorizontal,
                    textOptions: that._textOptions,
                    textFontStyles: that._textFontStyles,
                    textSpacing: options.label.minSpacing,
                    getText: getText,
                    renderText: function(text, x, y, options) {
                        return that._renderer.text(text, x, y, options).append(that._renderer.root)
                    },
                    translate: function(value, useAdditionalTranslator) {
                        return useAdditionalTranslator ? that._additionalTranslator.translate(value) : that._translator.translate(value)
                    },
                    addMinMax: options.showCustomBoundaryTicks ? that._boundaryTicksVisibility : void 0
                }
            },
            _getMinMax: function() {
                return {
                    min: this._options.min,
                    max: this._options.max
                }
            },
            _getStick: function() {
                return !this._options.valueMarginsEnabled
            },
            _getStripLabelCoords: function(stripLabelOptions, stripFrom, stripTo) {
                var x, y, that = this,
                    additionalTranslator = that._additionalTranslator,
                    isHorizontal = that._isHorizontal,
                    align = isHorizontal ? CENTER : LEFT;
                if (isHorizontal) {
                    if (stripLabelOptions.horizontalAlignment === CENTER) {
                        x = stripFrom + (stripTo - stripFrom) / 2;
                        align = CENTER
                    } else {
                        if (stripLabelOptions.horizontalAlignment === LEFT) {
                            x = stripFrom;
                            align = LEFT
                        } else {
                            if (stripLabelOptions.horizontalAlignment === RIGHT) {
                                x = stripTo;
                                align = RIGHT
                            }
                        }
                    }
                    y = additionalTranslator.translateSpecialCase(CANVAS_POSITION_PREFIX + stripLabelOptions.verticalAlignment)
                } else {
                    x = additionalTranslator.translateSpecialCase(CANVAS_POSITION_PREFIX + stripLabelOptions.horizontalAlignment);
                    align = stripLabelOptions.horizontalAlignment;
                    if (stripLabelOptions.verticalAlignment === TOP) {
                        y = stripFrom
                    } else {
                        if (stripLabelOptions.verticalAlignment === CENTER) {
                            y = stripTo + (stripFrom - stripTo) / 2
                        } else {
                            if (stripLabelOptions.verticalAlignment === BOTTOM) {
                                y = stripTo
                            }
                        }
                    }
                }
                return {
                    x: x,
                    y: y,
                    align: align
                }
            },
            _getTranslatedValue: function(value, y, offset) {
                return {
                    x: this._translator.translate(value, offset, "semidiscrete" === this._options.type && this._options.tickInterval),
                    y: y
                }
            },
            _getSkippedCategory: function() {
                var skippedCategory, categories = this._translator.getVisibleCategories() || this._translator.getBusinessRange().categories;
                if (categories && categories.length && !!this._tickOffset) {
                    skippedCategory = categories[categories.length - 1]
                }
                return skippedCategory
            },
            _getSpiderCategoryOption: $.noop
        }
    }
});
